// +---------------------------------------------------------------------------+
//
//      Program:    calibration_target_software.c
//
//      Purpose:    The software to run in the target (ATtiny4,5,9,10) during
//                  the oscillator calibration, based on application note AVR057
//
//      Author:     Herbert Dingfelder
//
//      License:    GPL 3
//
//      History:    04.04.2015  Initial Coding, doing all the mathematics to figure out
//                              which settings must be made to timers and which target
//                              values to aim for.
//                  06.04.2015  Added successive approximation with rounding, to find
//                              the best possible calibration value.
//                  07.04.2015  Added 1MHz out at start of programm, to be measured
//                              by an external counter to verify the functionality
//                  08.04.2015  Practical tests prove that the software works great
//                              => Version 1.0 released
//                              
//
// +---------------------------------------------------------------------------+

#define F_CPU 8000000
#include <util/delay.h>
#include <avr/io.h>

// ----- defines and makros for the timer handling -----

#define TPI_PORT PORTB
#define TPI_DDR  DDRB
#define TPI_IN   PINB
#define TPI_PU   PUEB
#define TPI_DATA _BV(0)
#define TPI_CLK  _BV(1)

#define LED_PORT PORTB
#define LED_DDR  DDRB
#define LED_BIT  _BV(2)

// switch on the counter with prescale factor 1, running at the system main clock of 8MHz
#define TIMER_START (TCCR0B = _BV(CS00))     
                            
// switch off the counter
#define TIMER_STOP (TCCR0B = 0x00)
    

// Define the target value for the calibration. 
// Calculation: 
// - Programmer runs at 11.059MHz. Compare match limit is set to 0xAD=173.
// - Output pin is set to toggle at compare match. That results in a reference
//   frequency from the programmer of 11.059MHz/(173+1)/2 = 31778.73Hz.
// - A reasonable number of the reference periods to measure is 80. That gives
//   around 20000 counts, which leaves a factor of 3 of margin for an overflow
//   of a 16bit counter variable. (Reasoning see below) => Per definition the
//   number of periods to count is set to 80.
// - At perfect calibration, the main clock frequency is 8MHz. At the given
//   reference frequency of 31.77873kHz, there will be 251.749 main clock 
//   cycles per reference periode. Within 80 reference periodes that gives
//   80 x 251.749 = 20139,2 as the target counter value.

#define NUMBER_OF_REF_PERIODS_TO_COUNT  80
#define TARGET_COUNTER_VALUE            20139 

void reset_handler(void);
void send_derived_value_to_programmer(uint8_t return_value);
uint16_t count_clock_cycles_for_ref_clock_periodes(uint8_t speed_setting_osccal);


// ============================== main starts here =============================
int main(void) 
{
    
    unsigned int trial_result, trial_higher;
    unsigned int base_deviation, next_deviation;
    unsigned int succ_appr_result=0;
    signed char succ_appr_bit_number=0;
    unsigned char succ_appr_trial_value=0;    

    // --------------------- main loop for numerous calibraion trials ---------------------
    for(;;)                                
    {
    
        reset_handler();                    // initialize main clock, GPIOs and counter
        
        // while waiting for the reference clock to be supplied, output a 1MHz signal on OC0A
        // (can be measured with an frequency counter, so see how precise the system clock is)
        
        TCCR0A = _BV(COM0A0);               // Toggle OC0A on compare match
                                            // (since O00A is the same as TPIDATA, it is already an output)
        TCCR0B = _BV(WGM02) | _BV(CS00);    // Clear Timer on Compare mode, counter on with prescaler=1
        OCR0A = 3;                          // f_out = 0.5*(f_mainclock/(OCR0A+1)) = 1MHz
        
        while( (TPI_IN & TPI_CLK)==0 );     // wait for TPICLK to go high
        while( (TPI_IN & TPI_CLK)!=0 );     // wait for falling edge of TPICLK
        
        TCCR0A = 0;                         // reset timer registers to default values
        TCCR0B = 0;
        OCR0A  = 0;
        
        // ----- the actual successive approximation of the perfect calibration value starts here -----
        
        // derive the result bit by bit, starting with the MSB, going down to LSB
        for( succ_appr_bit_number = 7 ; succ_appr_bit_number>=0 ; succ_appr_bit_number-- )
        {
            
            // the next value to try is the value, that we have so far, plus the value of the current bit to try
            succ_appr_trial_value = succ_appr_result | (1<<succ_appr_bit_number); 
            
            trial_result = count_clock_cycles_for_ref_clock_periodes(succ_appr_trial_value);     // try out what result we get with this value
        
            if( trial_result <= TARGET_COUNTER_VALUE )          // is the result within the range to our target value?
            {
                succ_appr_result = succ_appr_trial_value;       // if yes, the trial value was good, we use if further
            }

        }
        
        
        // now try if the derived result n gives the closest match or if n+1 would be closer (should we round upwards?)
        
        if( (succ_appr_result < 255 ) )  // does it make sense to try n+1 or are we on the limit already ?
        {
            
            trial_result = count_clock_cycles_for_ref_clock_periodes(succ_appr_result);          // try out both n and n+1
            trial_higher = count_clock_cycles_for_ref_clock_periodes(succ_appr_result+1);
            
            // since trial results can differ a bit from trial to trial (within a physical environment)
            //   we need to make sure that the sign of the difference calculation is correct
            
            if(trial_result>=TARGET_COUNTER_VALUE)              // calculate deviation with n
                base_deviation = trial_result - TARGET_COUNTER_VALUE;
            else
                base_deviation = TARGET_COUNTER_VALUE - trial_result;
        
            if(trial_higher>=TARGET_COUNTER_VALUE)              // calculate deviation with n+1
                next_deviation = trial_higher - TARGET_COUNTER_VALUE;
            else
                next_deviation = TARGET_COUNTER_VALUE - trial_higher;
                
            if(next_deviation < base_deviation)     // are we better off to use the next value, i.e round upwards?
            {
                succ_appr_result++;
            }
            
        }    
        
        OSCCAL = succ_appr_result;                              // adjust the oscillator to the ideal value
                                                                // that was found in the routine above, to
                                                                // have the 1MHz output running at that
                                                                // adjusted speed for external measurements
        
        TPI_PORT &=~TPI_DATA;                                   // take TPIDATA low to indicate to programmer
                                                                //   that calibration is complete
        
        send_derived_value_to_programmer( succ_appr_result );   // send the calibration value to programmer 
        
        
    }  
    // ----------------------------------- end main loop ---------------------------------
           
    return 0;

}


// -------------------- send derived value back to programmer ------------------
void send_derived_value_to_programmer(uint8_t return_value)
{
    
    uint8_t bitnumber;
    
    for(bitnumber=0; bitnumber<8; bitnumber++)  // send all 8 bits in a loop
    {
        while( (TPI_IN & TPI_CLK)==0 );     // wait for TPICLK to go high
        while( (TPI_IN & TPI_CLK)!=0 );     // wait for falling edge of TPICLK
        
        if( return_value & 0x80 )           // if MSB is set...
        {
            TPI_PORT |= TPI_DATA;           // put TPIDATA pin to high
        }
        else                                // if MSB is not set ...
        {
            TPI_PORT &= ~TPI_DATA;          // put TPIDATA pin to low
        }
        
        return_value <<= 1;                 // logic shift left, to move next bit to MSB position
    }
    
}

// ---- count the number of system clock cycles for the defined number of ref clock periodes ------
uint16_t count_clock_cycles_for_ref_clock_periodes(uint8_t speed_setting_osccal)
{
    
    uint8_t counter=0;
    uint16_t counting_result=0;
    
    OSCCAL = speed_setting_osccal;          // set the system clock adjustment to the new test value
    _delay_ms(1);                           // give the clocking system a moment to become stable
    
    while( (TPI_IN & TPI_CLK)==0 );         // wait for TPICLK to go high
    
    while( (TPI_IN & TPI_CLK)!=0 )          // wait for falling edge
    {
        LED_PORT &=~ LED_BIT;               // LED on (active low)
        LED_PORT |= LED_BIT;                // LED off (active low)
    }
    
    // we are ready to measure the number of main clock cycles during the 
    // the defined number of periodes of the reference clock frequency from the programmer
    
    LED_PORT &=~ LED_BIT;                   // LED on (active low)
    
    TCNT0 = 0;                              // reset counter (16bit register)
    TIMER_START;                            // start the stop-watch
    
    for( counter=0 ; counter<NUMBER_OF_REF_PERIODS_TO_COUNT ; counter++)
    {
        while( (TPI_IN & TPI_CLK)==0 );     // wait for TPICLK to go high
        while( (TPI_IN & TPI_CLK)!=0 );     // wait for falling edge
    }
    
    TIMER_STOP;                             // stop the stop-watch
    counting_result = TCNT0;                // save the counter state 
    
    LED_PORT |= LED_BIT;                    // LED off (active low)
    
    return counting_result;                 // hand back the number of system clock cycles during
                                            // the defined number of reference clock periodes

}
 

// --------------------------------- reset handler -----------------------------
void reset_handler(void)
{
    
    CCP = 0xD8;     // write signature to configuration change protection register, to allow system clock change
    CLKPSR = 0;     // switch main clock prescaler to factor 1, i.e. main clock runs at full 8 MHz 
                    // (default is 1:8 which lets the main clock run at 1MHz at startup)
    
    TPI_DDR  |= TPI_DATA;   // configure TPIDATA pin as output
    TPI_PORT |= TPI_DATA;   // pull the output high, to signal to programmer that calibration is not done yet
    
    TPI_PU  |= TPI_CLK;     // switch on pull-up resistor for TPICLK input    
     
    LED_DDR |= LED_BIT;     // for tracing during the calibration, use a LED on the remaining GPIO as output
    LED_PORT|= LED_BIT;     // LED is connected as active-low, pull signal high keep LED off
    
}



